import bpy
from mathutils import Vector
from ...utils import copy_bone, flip_bone, put_bone, org, align_bone_y_axis, align_bone_x_axis
from ...utils import strip_org, make_deformer_name, connected_children_names
from ...utils import create_circle_widget, create_sphere_widget, create_neck_bend_widget, create_neck_tweak_widget
from ..widgets import create_ballsocket_widget
from ...utils import MetarigError, make_mechanism_name, create_cube_widget
from rna_prop_ui import rna_idprop_ui_prop_get

script = """
controls = [%s]
torso    = '%s'

if is_selected( controls ):
    if hasattr(pose_bones[torso],'["%s"]'):
        layout.prop( pose_bones[ torso ], '["%s"]', slider = True )
    if hasattr(pose_bones[torso],'["%s"]'):
        layout.prop( pose_bones[ torso ], '["%s"]', slider = True )
    if hasattr(pose_bones[torso],'["%s"]'):
        layout.prop( pose_bones[ torso ], '["%s"]', slider = True )
"""


class Rig:

    def __init__(self, obj, bone_name, params):
        """ Initialize torso rig and key rig properties """

        eb = obj.data.edit_bones

        self.obj = obj
        self.org_bones = [bone_name] + connected_children_names(obj, bone_name)
        self.params = params
        # self.spine_length = sum([eb[b].length for b in self.org_bones])
        self.copy_rotation_axes = params.copy_rotation_axes
        self.use_head = params.use_head
        self.use_tail = params.use_tail

        # Check if user provided the pivot position
        if params.pivot_pos:
            self.pivot_pos = params.pivot_pos + 1
        else:
            raise MetarigError(
                "RIGIFY ERROR: please specify pivot bone position"
            )

        # Check if neck is lower than pivot
        if self.use_head and params.neck_pos <= params.pivot_pos and params.neck_pos != 0:
            raise MetarigError(
                "RIGIFY ERROR: Neck cannot be below or the same as pivot. (use 0 for no neck)"
            )
        else:
            self.neck_pos = params.neck_pos

        if not self.use_head:
            self.neck_pos = len(self.org_bones)

        if self.use_tail and self.pivot_pos - 2 > 0:
            self.tail_pos = params.tail_pos

        # Assign values to tweak layers props if opted by user
        if params.tweak_extra_layers:
            self.tweak_layers = list(params.tweak_layers)
        else:
            self.tweak_layers = None

        # Report error of user created less than the minimum of bones for rig
        min_bone_number = 3
        if self.use_head:
            min_bone_number += 1
        if self.use_tail:
            min_bone_number += 2

        if len(self.org_bones) < min_bone_number:
            raise MetarigError(
                "RIGIFY ERROR: invalid rig structure on %s" % (strip_org(bone_name))
            )

    def build_bone_structure(self):
        """ Divide meta-rig into lists of bones according to torso rig anatomy:
            Neck --> Upper torso --> Lower torso --> Tail (optional) """

        if self.pivot_pos and (self.neck_pos == 0 or self.neck_pos > self.pivot_pos):

            neck_index = self.neck_pos - 1
            pivot_index = self.pivot_pos - 1

            tail_index = 0
            if self.use_tail and self.tail_pos > 1:     # 2 bones for the tail at least
                tail_index = self.tail_pos - 1

            if self.use_head:
                neck_bones = self.org_bones[neck_index::]
                upper_torso_bones = self.org_bones[pivot_index :neck_index]
            else:
                neck_bones = []
                upper_torso_bones = self.org_bones[pivot_index ::]

            tail_bones = []
            if tail_index:
                lower_torso_bones = self.org_bones[tail_index + 1:pivot_index ]
                tail_bones = self.org_bones[:tail_index+1]
            else:
                lower_torso_bones = self.org_bones[:pivot_index ]

            torso_bones = upper_torso_bones + lower_torso_bones
            eb = self.obj.data.edit_bones
            self.spine_length = sum([eb[b].length for b in torso_bones])

            return {
                'neck': neck_bones,
                'upper': upper_torso_bones,
                'lower': lower_torso_bones,
                'tail': tail_bones
            }

        else:
            return 'ERROR'

    def orient_bone(self, eb, axis, scale, reverse=False):
        v = Vector((0,0,0))

        setattr(v, axis, scale)

        if reverse:
            tail_vec = v * self.obj.matrix_world
            eb.head[:] = eb.tail
            eb.tail[:] = eb.head + tail_vec
        else:
            tail_vec = v * self.obj.matrix_world
            eb.tail[:] = eb.head + tail_vec

    def create_pivot(self, pivot):
        """ Create the pivot control and mechanism bones """
        org_bones = self.org_bones
        pivot_name = org_bones[pivot-1]

        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones

        # Create torso control bone
        torso_name = 'torso'
        ctrl_name = copy_bone(self.obj, pivot_name, torso_name)
        ctrl_eb = eb[ctrl_name]

        self.orient_bone(ctrl_eb, 'y', self.spine_length * 0.6)

        # Create mch_pivot
        mch_name = make_mechanism_name('pivot')
        mch_name = copy_bone(self.obj, ctrl_name, mch_name)
        mch_eb = eb[mch_name]

        mch_eb.length /= 4

        # Positioning pivot in a more usable location for animators
        # if self.use_tail and self.tail_pos > 0:
        #     pivot_loc = eb[org_bones[pivot-1]].head
        if self.use_tail and self.tail_pos > 0:
            first_torso_bone = self.tail_pos
            pivot_loc = (eb[org_bones[first_torso_bone]].head + eb[org_bones[first_torso_bone]].tail)/2
        else:
            pivot_loc = (eb[org_bones[0]].head + eb[org_bones[0]].tail) / 2

        put_bone(self.obj, ctrl_name, pivot_loc)

        return {
            'ctrl': ctrl_name,
            'mch': mch_name
        }

    def create_deform(self):
        org_bones = self.org_bones

        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones

        def_bones = []
        for org_b in org_bones:
            def_name = make_deformer_name(strip_org(org_b))
            def_name = copy_bone(self.obj, org_b, def_name)
            def_bones.append(def_name)

        return def_bones

    def create_neck(self, neck_bones):
        org_bones = self.org_bones

        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones

        if not self.use_head:
            return {
                'ctrl_neck': '',
                'ctrl': '',
                'mch_str': '',
                'mch_neck': '',
                'mch_head': '',
                'mch': [],
                'tweak': [],
                'neck_bend': '',
                'original_names': neck_bones
            }

        neck, neck_bend = '', ''
        if len(neck_bones) >= 2:
            # Create neck control
            neck = copy_bone(self.obj, org(neck_bones[0]), 'neck')
            neck_eb = eb[neck]

            # Neck spans all neck bones (except head)
            neck_eb.tail[:] = eb[org(neck_bones[-1])].head

            if len(neck_bones) > 3:

                # Create neck bend control
                neck_bend = copy_bone(self.obj, org(neck_bones[0]), 'neck_bend')
                neck_bend_eb = eb[neck_bend]

                # Neck pivot position
                if (len(neck_bones)-1) % 2:     # odd num of neck bones (head excluded)
                    center_bone = org(neck_bones[int((len(neck_bones))/2) - 1])
                    neck_bend_eb.head = (eb[center_bone].head + eb[center_bone].tail)/2
                else:
                    center_bone = org(neck_bones[int((len(neck_bones)-1)/2) - 1])
                    neck_bend_eb.head = eb[center_bone].tail

                align_bone_y_axis(self.obj, neck_bend, eb[neck].y_axis)
                align_bone_x_axis(self.obj, neck_bend, eb[neck].x_axis)
                eb[neck_bend].length = eb[neck].length / 2

        # Create head control
        head = copy_bone(self.obj, org(neck_bones[-1]), 'head')

        # MCH bones
        mch_str, mch_neck = '', ''
        if len(neck_bones) >= 2:
            # Neck MCH stretch
            mch_str = copy_bone(self.obj, neck, make_mechanism_name('STR-neck'))

            # Neck MCH rotation
            mch_neck = copy_bone(
                self.obj, neck, make_mechanism_name('ROT-neck')
            )

            self.orient_bone(eb[mch_neck], 'y', self.spine_length / 10)

        # Head MCH rotation
        mch_head = copy_bone(self.obj, head, make_mechanism_name('ROT-head'))
        self.orient_bone(eb[mch_head], 'y', self.spine_length / 10)

        twk, mch = [], []

        if len(neck_bones) >= 2:
            # Intermediary bones
            for b in neck_bones[1:-1]:  # All except 1st (neck) and last (head)
                mch_name = copy_bone(self.obj, org(b), make_mechanism_name(b))
                eb[mch_name].length /= 4
                align_bone_y_axis(self.obj, mch_name, eb[neck].y_axis)
                align_bone_x_axis(self.obj, mch_name, eb[neck].x_axis)
                eb[mch_name].use_inherit_scale = False
                mch += [mch_name]

            # Tweak bones
            for b in neck_bones[:-1]:   # All except last bone
                twk_name = "tweak_" + b
                twk_name = copy_bone(self.obj, org(b), twk_name)

                eb[twk_name].length /= 2

                twk += [twk_name]

        return {
            'ctrl_neck': neck,
            'ctrl': head,
            'mch_str': mch_str,
            'mch_neck': mch_neck,
            'mch_head': mch_head,
            'mch': mch,
            'tweak': twk,
            'neck_bend': neck_bend,
            'original_names': neck_bones
        }

    def create_chest(self, chest_bones):
        org_bones = self.org_bones

        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones

        # get total spine length

        # Create chest control bone
        chest = copy_bone(self.obj, org(chest_bones[0]), 'chest')
        self.orient_bone(eb[chest], 'y', self.spine_length / 3)

        # create chest mch_wgt
        mch_wgt = copy_bone(
            self.obj, org(chest_bones[-1]),
            make_mechanism_name('WGT-chest')
        )

        # Create mch and twk bones
        twk, mch = [], []

        for b in chest_bones:
            mch_name = copy_bone( self.obj, org(b), make_mechanism_name(b) )
            self.orient_bone( eb[mch_name], 'y', self.spine_length / 10 )

            twk_name = "tweak_" + b
            twk_name = copy_bone( self.obj, org(b), twk_name )
            eb[twk_name].length /= 2

            mch += [ mch_name ]
            twk += [ twk_name ]

        return {
            'ctrl'    : chest,
            'mch'     : mch,
            'tweak'   : twk,
            'mch_wgt' : mch_wgt
        }

    def create_hips(self, hip_bones):
        org_bones = self.org_bones

        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones

        # Create hips control bone
        hips = copy_bone(self.obj, org(hip_bones[-1]), 'hips')
        self.orient_bone(
            eb[hips],
            'y',
            self.spine_length / 4,
            reverse = True
        )

        # create hips mch_wgt
        mch_wgt = copy_bone(
            self.obj, org(hip_bones[0]),
            make_mechanism_name('WGT-hips')
        )

        # Create mch and tweak bones
        twk, mch = [], []
        for b in hip_bones:
            mch_name = copy_bone( self.obj, org(b), make_mechanism_name(b) )
            self.orient_bone(
                eb[mch_name], 'y', self.spine_length / 10, reverse = True
            )

            twk_name = "tweak_" + b
            twk_name = copy_bone( self.obj, org(b), twk_name )

            eb[twk_name].length /= 2

            mch += [ mch_name ]
            twk += [ twk_name ]

        return {
            'ctrl'    : hips,
            'mch'     : mch,
            'tweak'   : twk,
            'mch_wgt' : mch_wgt
        }

    def create_tail(self, tail_bones):
        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones
        org_bones = self.org_bones

        ctrl_chain = []
        for i in range(len(tail_bones)):
            name = tail_bones[i]

            ctrl_bone = copy_bone(
                self.obj,
                org(name),
                strip_org(name)
            )

            flip_bone(self.obj, ctrl_bone)
            ctrl_chain.append(ctrl_bone)

        # Main ctrl
        name = tail_bones[-1]
        main_ctrl_bone = copy_bone(
            self.obj,
            org(name),
            strip_org(name).split('.')[0] + "_master"
        )
        flip_bone(self.obj, main_ctrl_bone)

        mch_tail = ''
        tail_first = org_bones[self.tail_pos-1]
        mch_rot_tail = copy_bone(
                self.obj,
                org(tail_first),
                make_mechanism_name("ROT-tail")
            )

        self.orient_bone(eb[mch_rot_tail], 'y', eb[tail_first].length)
        put_bone(self.obj, mch_rot_tail, eb[tail_first].tail)
        mch_tail = mch_rot_tail

        tweak_chain = []
        for i in range(len(tail_bones)):
            name = tail_bones[i]

            tweak_bone = copy_bone(
                self.obj,
                org(name),
                "tweak_" + strip_org(name)
            )

            tweak_e = eb[tweak_bone]
            tweak_e.length /= 2  # Set size to half

            # Position tweaks
            flip_bone(self.obj, tweak_bone)
            put_bone(self.obj, tweak_bone, eb[org(name)].head)

            tweak_chain.append(tweak_bone)

        return {
            'ctrl': ctrl_chain,
            'ctrl_tail': main_ctrl_bone,
            'mch_tail': mch_tail,
            'tweak': tweak_chain,
            'original_names': tail_bones
        }

    def parent_bones(self, bones):
        org_bones = self.org_bones
        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones

        # Parent deform bones
        for i, b in enumerate(bones['def']):
            if i > 0:   # For all bones but the first (which has no parent)
                eb[b].parent = eb[bones['def'][i-1]]    # to previous
                eb[b].use_connect = True

        # Parent control bones
        # Head control => MCH-rotation_head
        if self.use_head:
            eb[bones['neck']['ctrl']].parent = eb[bones['neck']['mch_head']]

        # Tail control chain
        if self.use_tail:
            tail_ctrl = bones['tail']['ctrl']
            for i, b in enumerate(tail_ctrl[:-1]):
                eb[b].parent = eb[tail_ctrl[i+1]]
            eb[tail_ctrl[-1]].parent = eb[bones['tail']['mch_tail']]
            eb[bones['tail']['ctrl_tail']].parent = eb[org_bones[self.tail_pos]]

        if bones['neck']['ctrl_neck']:
            # MCH stretch => neck ctrl
            eb[ bones['neck']['mch_str']].parent = eb[ bones['neck']['ctrl_neck']]

            # Neck control => MCH-rotation_neck
            eb[bones['neck']['ctrl_neck']].parent = eb[bones['neck']['mch_neck']]

            # Neck pivot => MCH-rotation_neck
            if bones['neck']['neck_bend']:
                # eb[bones['neck']['neck_bend']].parent = eb[bones['neck']['ctrl_neck']]
                eb[bones['neck']['neck_bend']].parent = eb[bones['neck']['mch_str']]

        # Parent hips and chest controls to torso
        eb[bones['chest']['ctrl']].parent = eb[bones['pivot']['ctrl']]
        eb[bones['hips']['ctrl']].parent = eb[bones['pivot']['ctrl']]

        # Parent mch bones
        if bones['neck']['ctrl_neck']:
            # Neck mch
            eb[bones['neck']['mch_head']].parent = eb[bones['neck']['ctrl_neck']]
        elif self.use_head:
            eb[bones['neck']['mch_head']].parent = eb[bones['chest']['mch'][-1]]

        for i, b in enumerate([eb[n] for n in bones['neck']['mch']]):
            b.parent = eb[bones['neck']['mch_str']]
            # for org_b in bones['neck']['original_names']:
            #     if org_b in b.name:
            #         b.parent = eb[org(org_b)]

        # Chest mch bones and neck mch
        chest_mch = bones['chest']['mch'] + [bones['neck']['mch_neck']]
        for i, b in enumerate(chest_mch):
            if i == 0:
                eb[b].parent = eb[bones['pivot']['ctrl']]
            elif b:
                eb[b].parent = eb[chest_mch[i-1]]

        # Hips mch bones
        for i, b in enumerate(bones['hips']['mch']):
            if i == len(bones['hips']['mch']) - 1:
                eb[b].parent = eb[bones['pivot']['ctrl']]
            else:
                eb[b].parent = eb[bones['hips']['mch'][i+1]]

        # mch pivot
        eb[bones['pivot']['mch']].parent = eb[bones['chest']['mch'][0]]

        # MCH widgets
        eb[bones['chest']['mch_wgt']].parent = eb[bones['chest']['mch'][-1]]
        eb[bones['hips']['mch_wgt']].parent = eb[bones['hips']['mch'][0]]

        # Neck Tweaks
        if bones['neck']['tweak']:
            # Neck tweaks
            for i, twk in enumerate( bones['neck']['tweak']):
                if i == 0:
                    eb[twk].parent = eb[ bones['neck']['ctrl_neck']]
                else:
                    eb[twk].parent = eb[ bones['neck']['mch'][i-1]]

        # Chest tweaks
        for twk, mch in zip( bones['chest']['tweak'], bones['chest']['mch']):
            if bones['chest']['tweak'].index(twk) == 0:
                eb[twk].parent = eb[bones['pivot']['mch']]
            else:
                eb[twk].parent = eb[mch]

        # Hips tweaks
        for i, twk in enumerate(bones['hips']['tweak']):
            if i == 0:
                eb[twk].parent = eb[bones['hips']['mch'][i]]
            else:
                eb[twk].parent = eb[bones['hips']['mch'][i-1]]

        # Tail mchs
        if self.use_tail:
            mch_rot_tail = bones['tail']['mch_tail']
            eb[mch_rot_tail].parent = eb[bones['hips']['tweak'][0]]

        # Tail tweaks
        if self.use_tail:
            for i, twk in enumerate(bones['tail']['tweak']):
                if i == 0:
                    eb[twk].parent = eb[bones['tail']['ctrl'][i]]
                else:
                    eb[twk].parent = eb[bones['tail']['ctrl'][i-1]]

        # Parent orgs to matching tweaks
        tweaks = []
        if self.use_tail:
            tweaks += bones['tail']['tweak']

        tweaks += bones['hips']['tweak'] + bones['chest']['tweak']
        if self.use_head:
            tweaks += bones['neck']['tweak'] + [bones['neck']['ctrl']]

        original_neck_bones = [org(b) for b in bones['neck']['original_names']]
        original_neck_bones = original_neck_bones[1:-1]  # exclude first neck bone and head
        for b, twk in zip(org_bones[:-1], tweaks):
            if b in original_neck_bones and len(bones['neck']['original_names']) > 3:
                idx = org_bones.index(b)
                org_parent = org_bones[idx-1]
                eb[b].parent = eb[org_parent]
            else:
                eb[b].parent = eb[twk]

        if self.use_head:
            eb[org_bones[-1]].parent = eb[bones['neck']['ctrl']]

    def make_constraint(self, bone, constraint):
        bpy.ops.object.mode_set(mode='OBJECT')
        pb = self.obj.pose.bones

        owner_pb = pb[bone]
        const = owner_pb.constraints.new(constraint['constraint'])
        const.target = self.obj

        # filter contraint props to those that actually exist in the currnet
        # type of constraint, then assign values to each
        for p in [k for k in constraint.keys() if k in dir(const)]:
            setattr(const, p, constraint[p])

    def constrain_bones(self, bones):
        # MCH bones

        # head and neck MCH bones
        for b in [bones['neck']['mch_head'], bones['neck']['mch_neck']]:
            if b:
                self.make_constraint(b, {
                    'constraint': 'COPY_ROTATION',
                    'subtarget': bones['pivot']['ctrl'],
                })
                self.make_constraint(b, {
                    'constraint': 'COPY_SCALE',
                    'subtarget': bones['pivot']['ctrl'],
                })

        if bones['neck']['mch_str']:
            # Neck MCH Stretch
            self.make_constraint(bones['neck']['mch_str'], {
                'constraint': 'DAMPED_TRACK',
                'subtarget': bones['neck']['ctrl'],
            })
            self.make_constraint(bones['neck']['mch_str'], {
                'constraint': 'STRETCH_TO',
                'subtarget': bones['neck']['ctrl'],
            })

        # Intermediary mch bones
        intermediaries = [bones['neck'], bones['chest'], bones['hips']]

        for i, l in enumerate(intermediaries):
            mch = l['mch']

            for j, b in enumerate(mch):

                if i == 0:      # Neck mch-s
                    if len(bones['neck']['original_names']) > 3:
                        self.make_constraint(b, {
                            'constraint': 'COPY_LOCATION',
                            'subtarget': org(l['original_names'][j+1]),
                            'influence': 1.0
                        })
                    else:
                        nfactor = float((j + 1) / len(mch))
                        self.make_constraint(b, {
                            'constraint': 'COPY_ROTATION',
                            'subtarget': l['ctrl'],
                            'influence': nfactor
                        })

                    step = 2/(len(mch)+1)
                    xval = (j+1)*step
                    influence = 2*xval - xval**2    #parabolic influence of pivot

                    if bones['neck']['neck_bend']:
                        self.make_constraint(b, {
                            'constraint': 'COPY_LOCATION',
                            'subtarget': l['neck_bend'],
                            'influence': influence,
                            'use_offset': True,
                            'owner_space': 'LOCAL',
                            'target_space': 'LOCAL'
                        })

                    if len(bones['neck']['original_names']) > 3:
                        self.make_constraint(b, {
                            'constraint': 'COPY_SCALE',
                            'subtarget': org(l['original_names'][j+1]),
                            'influence': 1.0
                        })

                else:
                    factor = float(1 / len(l['tweak']))
                    self.make_constraint(b, {
                        'constraint': 'COPY_TRANSFORMS',
                        'subtarget': l['ctrl'],
                        'influence': factor,
                        'owner_space': 'LOCAL',
                        'target_space': 'LOCAL'
                    })

        # Tail ctrls
        if self.use_tail:
            tail_ctrl = bones['tail']['ctrl']
            tail_ctrl.append(bones['tail']['ctrl_tail'])

            for i, b in enumerate(tail_ctrl[:-1]):
                self.make_constraint(b, {
                    'constraint': 'COPY_ROTATION',
                    'subtarget': tail_ctrl[i+1],
                    'influence': 1.0,
                    'use_x': self.copy_rotation_axes[0],
                    'use_y': self.copy_rotation_axes[1],
                    'use_z': self.copy_rotation_axes[2],
                    'use_offset': True,
                    'owner_space': 'LOCAL',
                    'target_space': 'LOCAL'
                })

            b = bones['tail']['mch_tail']
            self.make_constraint(b, {
                'constraint': 'COPY_ROTATION',
                'subtarget': bones['pivot']['ctrl'],
                'influence': 1.0,
            })

        # MCH pivot
        self.make_constraint(bones['pivot']['mch'], {
            'constraint': 'COPY_TRANSFORMS',
            'subtarget': bones['hips']['mch'][-1],
            'owner_space': 'LOCAL',
            'target_space': 'LOCAL'
        })

        # DEF bones
        deform = bones['def']
        tweaks = []
        if self.use_tail:
            tweaks += bones['tail']['tweak']

        tweaks += bones['hips']['tweak'] + bones['chest']['tweak']
        if self.use_head:
            tweaks += bones['neck']['tweak'] + [bones['neck']['ctrl']]

        for d, t in zip(deform, tweaks):
            tidx = tweaks.index(t)

            self.make_constraint(d, {
                'constraint': 'COPY_TRANSFORMS',
                'subtarget': t
            })

            if tidx != len(tweaks) - 1:
                if self.use_tail and t in bones['tail']['tweak']:
                    self.make_constraint(d, {
                        'constraint': 'DAMPED_TRACK',
                        'subtarget': tweaks[tidx + 1],
                        'track_axis': 'TRACK_NEGATIVE_Y'
                    })
                else:
                    self.make_constraint(d, {
                        'constraint': 'DAMPED_TRACK',
                        'subtarget': tweaks[tidx + 1],
                    })

                self.make_constraint(d, {
                    'constraint': 'STRETCH_TO',
                    'subtarget': tweaks[tidx + 1],
                })

        pb = self.obj.pose.bones

        if bones['neck']['neck_bend']:
            pb[bones['neck']['neck_bend']].rotation_mode = 'ZXY'
            pb[bones['neck']['neck_bend']].lock_rotation[0] = True
            pb[bones['neck']['neck_bend']].lock_rotation[2] = True

        for t in tweaks:
            if t != bones['neck']['ctrl']:
                pb[t].rotation_mode = 'ZXY'

        original_neck_bones = [org(b) for b in bones['neck']['original_names']]
        # make IK on neck ORGs
        if len(original_neck_bones) > 3:
            last_neck = original_neck_bones[-2]
            self.make_constraint(last_neck, {
                'constraint': 'IK',
                'subtarget': bones['neck']['ctrl'],
                'chain_count': len(original_neck_bones) - 1
            })

            for b in original_neck_bones[:-1]:
                pb[b].ik_stretch = 0.1

    def create_drivers(self, bones):
        bpy.ops.object.mode_set(mode='OBJECT')
        pb = self.obj.pose.bones

        # Setting the torso's props
        torso = pb[bones['pivot']['ctrl']]

        props = []
        owners = []

        if self.use_head:
            props += ["head_follow"]
            owners += [bones['neck']['mch_head']]
            if bones['neck']['mch_neck']:
                props += ["neck_follow"]
                owners += [bones['neck']['mch_neck']]
        if self.use_tail:
            props += ["tail_follow"]
            owners += [bones['tail']['mch_tail']]

        for prop in props:
            if prop == 'neck_follow':
                torso[prop] = 0.5
            else:
                torso[prop] = 0.0

            prop = rna_idprop_ui_prop_get(torso, prop, create=True)
            prop["min"] = 0.0
            prop["max"] = 1.0
            prop["soft_min"] = 0.0
            prop["soft_max"] = 1.0
            prop["description"] = prop

        # driving the follow rotation switches for neck and head
        for bone, prop, in zip(owners, props):
            # Add driver to copy rotation constraint
            drv = pb[bone].constraints[0].driver_add("influence").driver
            drv.type = 'AVERAGE'

            var = drv.variables.new()
            var.name = prop
            var.type = "SINGLE_PROP"
            var.targets[0].id = self.obj
            var.targets[0].data_path = \
                torso.path_from_id() + '[' + '"' + prop + '"' + ']'

            drv_modifier = self.obj.animation_data.drivers[-1].modifiers[0]

            drv_modifier.mode = 'POLYNOMIAL'
            drv_modifier.poly_order = 1
            drv_modifier.coefficients[0] = 1.0
            drv_modifier.coefficients[1] = -1.0

    def locks_and_widgets(self, bones):
        bpy.ops.object.mode_set(mode='OBJECT')
        pb = self.obj.pose.bones

        # deform bones bbone segements
        for bone in bones['def'][:-1]:
            self.obj.data.bones[bone].bbone_segments = 8

        self.obj.data.bones[bones['def'][0]].bbone_in = 0.0
        self.obj.data.bones[bones['def'][-2]].bbone_out = 1.0

        # Locks
        tweaks = bones['neck']['tweak'] + bones['chest']['tweak']
        tweaks += bones['hips']['tweak']

        if self.use_tail:
            tweaks += bones['tail']['tweak']
            pb[bones['tail']['ctrl_tail']].lock_location = True, True, True

        # Tweak bones locks
        for bone in tweaks:
            pb[bone].lock_rotation = True, False, True
            pb[bone].lock_scale = False, True, False

        # Widgets

        # Assigning a widget to torso bone
        create_cube_widget(
            self.obj,
            bones['pivot']['ctrl'],
            radius=0.5,
            bone_transform_name=None
        )

        # Assigning widgets to control bones
        gen_ctrls = [
            bones['chest']['ctrl'],
            bones['hips']['ctrl']
        ]

        tail_ctrls = []
        if self.use_tail and bones['tail']['ctrl']:
            tail_ctrls = bones['tail']['ctrl'] + [bones['tail']['ctrl_tail']]
            gen_ctrls.extend(bones['tail']['ctrl'])

            create_ballsocket_widget(
                self.obj,
                bones['tail']['ctrl_tail'],
                size=0.7,
                bone_transform_name=None
            )

        for bone in gen_ctrls:

            if bone in tail_ctrls:
                radius = 0.5
            else:
                radius = 1.0

            # place chest on neck-base for very long necks
            if bone == bones['chest']['ctrl'] and len(bones['neck']['original_names']) > 3:
                head_tail = 0.0
            else:
                head_tail = 0.75

            create_circle_widget(
                self.obj,
                bone,
                radius=radius,
                head_tail=head_tail,
                with_line=False,
                bone_transform_name=None
            )

        if bones['neck']['ctrl_neck']:
            # Neck ctrl widget
            if len(bones['neck']['mch']) == 0:
                radius = 1
            else:
                radius = 1/(len(bones['neck']['mch']))
            create_circle_widget(
                self.obj,
                bones['neck']['ctrl_neck'],
                radius=radius,
                head_tail=0.5,
                bone_transform_name=None
            )

        if bones['neck']['neck_bend']:
            # Neck pivot widget
            if len(bones['neck']['mch']) == 0:
                radius = 0.5
            else:
                radius = 1/(2*len(bones['neck']['mch']))
            create_neck_bend_widget(
                self.obj,
                bones['neck']['neck_bend'],
                radius=radius,
                head_tail=0.0,
                bone_transform_name=None
            )

        # Head widget
        # place wgt @ middle of head bone for long necks
        if len(bones['neck']['original_names']) > 3:
            head_tail = 0.5
        else:
            head_tail = 1.0
        if self.use_head:
            create_circle_widget(
                self.obj,
                bones['neck']['ctrl'],
                radius              = 0.5,
                head_tail           = head_tail,
                with_line           = False,
                bone_transform_name = None
            )

        # place widgets on correct bones
        chest_widget_loc = pb[bones['chest']['mch_wgt']]
        pb[bones['chest']['ctrl']].custom_shape_transform = chest_widget_loc

        hips_widget_loc = pb[bones['hips']['mch_wgt']]

        if self.use_tail:
            hips_widget_loc = pb[bones['def'][self.tail_pos]]
            pb[bones['tail']['ctrl_tail']].custom_shape_transform = pb[bones['tail']['tweak'][0]]

        pb[bones['hips']['ctrl']].custom_shape_transform = hips_widget_loc

        # Assigning widgets to tweak bones and layers
        for bone in tweaks:

            if bones['neck']['tweak'] and bone == bones['neck']['tweak'][0] \
                    and len(bones['neck']['original_names']) > 3:
                create_neck_tweak_widget(self.obj, bone, size=1.0, bone_transform_name=None)
                continue
            create_sphere_widget(self.obj, bone, bone_transform_name=None)

            if self.tweak_layers:
                pb[bone].bone.layers = self.tweak_layers

    def generate(self):
        # Torso Rig Anatomy:
        # Neck: all bones above neck point, last bone is head
        # Upper torso: all bones between pivot and neck start
        # Lower torso: all bones below pivot until tail point
        # Tail: all bones below tail point

        bone_chains = self.build_bone_structure()

        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones

        # Clear parents for org bones
        for bone in self.org_bones:
            eb[bone].use_connect = False
            eb[bone].parent = None

        if bone_chains != 'ERROR':

            # Create lists of bones and strip "ORG" from their names
            neck_bones = [strip_org(b) for b in bone_chains['neck']]
            upper_torso_bones = [strip_org(b) for b in bone_chains['upper']]
            lower_torso_bones = [strip_org(b) for b in bone_chains['lower']]
            tail_bones = [strip_org(b) for b in bone_chains['tail']]

            bones = {}

            bones['def'] = self.create_deform()     # Gets org bones from self
            bones['pivot'] = self.create_pivot(self.pivot_pos)
            bones['neck'] = self.create_neck(neck_bones)
            bones['chest'] = self.create_chest(upper_torso_bones)
            bones['hips'] = self.create_hips(lower_torso_bones)

            # TODO: Add create tail
            if tail_bones:
                bones['tail'] = self.create_tail(tail_bones)

            # TEST
            bpy.ops.object.mode_set(mode='EDIT')
            eb = self.obj.data.edit_bones

            self.parent_bones(bones)
            self.constrain_bones(bones)
            self.create_drivers(bones)
            self.locks_and_widgets(bones)

        else:
            return

        controls = [bones['neck']['ctrl'],  bones['neck']['ctrl_neck']]
        controls += [bones['chest']['ctrl'], bones['hips']['ctrl']]
        controls += [bones['pivot']['ctrl']]

        if self.use_tail:
            controls.extend(bones['tail']['ctrl'])

        # Create UI
        controls_string = ", ".join(["'" + x + "'" for x in controls])
        return [script % (
            controls_string,
            bones['pivot']['ctrl'],
            'head_follow',
            'head_follow',
            'neck_follow',
            'neck_follow',
            'tail_follow',
            'tail_follow',
            )]


def add_parameters(params):
    """ Add the parameters of this rig type to the
        RigifyParameters PropertyGroup
    """
    params.neck_pos = bpy.props.IntProperty(
        name        = 'neck_position',
        default     = 6,
        min         = 0,
        description = 'Neck start position'
    )

    params.pivot_pos = bpy.props.IntProperty(
        name='pivot_position',
        default=2,
        min=0,
        description='Position of the torso control and pivot point'
    )

    params.copy_rotation_axes = bpy.props.BoolVectorProperty(
        size=3,
        description="Automation axes",
        default=tuple([i == 0 for i in range(0, 3)])
    )

    params.tail_pos = bpy.props.IntProperty(
        name='tail_position',
        default=2,
        min=2,
        description='Where the tail starts'
    )

    params.use_tail = bpy.props.BoolProperty(
        name='use_tail',
        default=False,
        description='Create tail bones'
    )

    params.use_head = bpy.props.BoolProperty(
        name='use_head',
        default=True,
        description='Create head and neck bones'
    )

    # Setting up extra layers for the FK and tweak
    params.tweak_extra_layers = bpy.props.BoolProperty(
        name="tweak_extra_layers",
        default=True,
        description=""
        )

    params.tweak_layers = bpy.props.BoolVectorProperty(
        size        = 32,
        description = "Layers for the tweak controls to be on",
        default     = tuple( [ i == 1 for i in range(0, 32) ] )
        )


def parameters_ui(layout, params):
    """ Create the ui for the rig parameters."""

    r = layout.row(align=True)
    r.prop(params, "use_head", toggle=True, text="Head")
    r.prop(params, "use_tail", toggle=True, text="Tail")

    r = layout.row()
    r.prop(params, "neck_pos")
    r.enabled = params.use_head

    r = layout.row()
    r.prop(params, "pivot_pos")

    r = layout.row()
    r.prop(params, "tail_pos")
    r.enabled = params.use_tail

    r = layout.row()
    col = r.column(align=True)
    row = col.row(align=True)
    for i, axis in enumerate(['x', 'y', 'z']):
        row.prop(params, "copy_rotation_axes", index=i, toggle=True, text=axis)
    r.enabled = params.use_tail

    r = layout.row()
    r.prop(params, "tweak_extra_layers")
    r.active = params.tweak_extra_layers

    col = r.column(align=True)
    row = col.row(align=True)

    bone_layers = bpy.context.active_pose_bone.bone.layers[:]

    for i in range(8):
        icon = "NONE"
        if bone_layers[i]:
            icon = "LAYER_ACTIVE"
        row.prop(params, "tweak_layers", index=i, toggle=True, text="", icon=icon)

    row = col.row(align=True)

    for i in range(16, 24):
        icon = "NONE"
        if bone_layers[i]:
            icon = "LAYER_ACTIVE"
        row.prop(params, "tweak_layers", index=i, toggle=True, text="", icon=icon)

    col = r.column(align=True)
    row = col.row(align=True)

    for i in range(8, 16):
        icon = "NONE"
        if bone_layers[i]:
            icon = "LAYER_ACTIVE"
        row.prop(params, "tweak_layers", index=i, toggle=True, text="", icon=icon)

    row = col.row(align=True)

    for i in range(24, 32):
        icon = "NONE"
        if bone_layers[i]:
            icon = "LAYER_ACTIVE"
        row.prop(params, "tweak_layers", index=i, toggle=True, text="", icon=icon)


def create_sample(obj):
    # generated by rigify.utils.write_metarig
    bpy.ops.object.mode_set(mode='EDIT')
    arm = obj.data

    bones = {}

    bone = arm.edit_bones.new('spine')
    bone.head[:] = 0.0000, 0.0552, 1.0099
    bone.tail[:] = 0.0000, 0.0172, 1.1573
    bone.roll = 0.0000
    bone.use_connect = False
    bones['spine'] = bone.name

    bone = arm.edit_bones.new('spine.001')
    bone.head[:] = 0.0000, 0.0172, 1.1573
    bone.tail[:] = 0.0000, 0.0004, 1.2929
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['spine']]
    bones['spine.001'] = bone.name

    bone = arm.edit_bones.new('spine.002')
    bone.head[:] = 0.0000, 0.0004, 1.2929
    bone.tail[:] = 0.0000, 0.0059, 1.4657
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['spine.001']]
    bones['spine.002'] = bone.name

    bone = arm.edit_bones.new('spine.003')
    bone.head[:] = 0.0000, 0.0059, 1.4657
    bone.tail[:] = 0.0000, 0.0114, 1.6582
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['spine.002']]
    bones['spine.003'] = bone.name

    bone = arm.edit_bones.new('spine.004')
    bone.head[:] = 0.0000, 0.0114, 1.6582
    bone.tail[:] = 0.0000, -0.013, 1.7197
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['spine.003']]
    bones['spine.004'] = bone.name

    bone = arm.edit_bones.new('spine.005')
    bone.head[:] = 0.0000, -0.013, 1.7197
    bone.tail[:] = 0.0000, -0.0247, 1.7813
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['spine.004']]
    bones['spine.005'] = bone.name

    bone = arm.edit_bones.new('spine.006')
    bone.head[:] = 0.0000, -0.0247, 1.7813
    bone.tail[:] = 0.0000, -0.0247, 1.9796
    bone.roll = 0.0000
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['spine.005']]
    bones['spine.006'] = bone.name


    bpy.ops.object.mode_set(mode='OBJECT')
    pbone = obj.pose.bones[bones['spine']]
    pbone.rigify_type = 'spines.super_spine'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'

    try:
        pbone.rigify_parameters.neck_pos = 5
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.tweak_layers = [False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['spine.001']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['spine.002']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['spine.003']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['spine.004']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['spine.005']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['spine.006']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'

    bpy.ops.object.mode_set(mode='EDIT')
    for bone in arm.edit_bones:
        bone.select = False
        bone.select_head = False
        bone.select_tail = False
    for b in bones:
        bone = arm.edit_bones[bones[b]]
        bone.select = True
        bone.select_head = True
        bone.select_tail = True
        arm.edit_bones.active = bone



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 3
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8-80 compliant>

# This script was developed with financial support from the Foundation for
# Science and Technology of Portugal, under the grant SFRH/BD/66452/2009.


bl_info = {
    "name": "C3D Graphics Lab Motion Capture file (.c3d)",
    "author": "Daniel Monteiro Basso <daniel@basso.inf.br>",
    "version": (2015, 5, 5, 1),
    "blender": (2, 74, 1),
    "location": "File > Import",
    "description": "Imports C3D Graphics Lab Motion Capture files",
    "wiki_url": "http://wiki.blender.org/index.php/Extensions:2.6/Py/"
                "Scripts/Import-Export/C3D_Importer",
    "category": 'Import-Export',
}


import bpy
from bpy.props import (
    StringProperty,
    BoolProperty,
    FloatProperty,
    IntProperty,
)

import os
import math
from mathutils import Vector
from . import import_c3d


class C3DAnimateCloud(bpy.types.Operator):
    """
        Animate the Marker Cloud
    """
    bl_idname = "import_anim.c3danim"
    bl_label = "Animate C3D"

    is_armature = False
    markerset = None
    uname = None
    curframe = 0
    fskip = 0
    scale = 0
    timer = None
    Y_up = False

    def update_empty(self, fno, ml, m):
        name = self.unames[self.prefix + ml]
        o = bpy.context.scene.objects[name]
        p = Vector(m.position) * self.scale
        o.location = Vector((p[0], -p[2], p[1])) if self.Y_up else p
        o.keyframe_insert('location', frame=fno)

    def update_bone(self, fno, ml, m, bones):
        name = self.prefix + ml
        if name not in bones:
            return
        b = bones[name]
        p = Vector(m.position) * self.scale
        b.matrix.translation = Vector((p[0], -p[2], p[1])) if self.Y_up else p
        b.keyframe_insert('location', -1, fno, name)

    def update_frame(self):
        fno = self.curframe
        if not self.use_frame_no:
            fno = (self.curframe - self.markerset.startFrame) / self.fskip
        for i in range(self.fskip):
            self.markerset.readNextFrameData()
        if self.is_armature:
            bones = bpy.context.active_object.pose.bones
        for ml in self.markerset.markerLabels:
            m = self.markerset.getMarker(ml, self.curframe)
            if m.confidence < self.confidence:
                continue
            if self.is_armature:
                self.update_bone(fno, ml, m, bones)
            else:
                self.update_empty(fno, ml, m)

    def modal(self, context, event):
        if event.type == 'ESC':
            return self.cancel(context)
        if event.type == 'TIMER':
            if self.curframe > self.markerset.endFrame:
                return self.cancel(context)
            self.update_frame()
            self.curframe += self.fskip
        return {'PASS_THROUGH'}

    def execute(self, context):
        self.timer = context.window_manager.\
            event_timer_add(0.001, context.window)
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

    def cancel(self, context):
        bpy.context.scene.frame_set(bpy.context.scene.frame_current)
        context.window_manager.event_timer_remove(self.timer)
        return {'FINISHED'}


class C3DImporter(bpy.types.Operator):
    """
        Load a C3D Marker Cloud
    """
    bl_idname = "import_anim.c3d"
    bl_label = "Import C3D"

    filepath = StringProperty(
        subtype='FILE_PATH',
    )
    Y_up = BoolProperty(
        name="Up vector is Y axis",
        default=False,
        description="Check when the data uses Y-up, uncheck when it uses Z-up",
    )
    from_inches = BoolProperty(
        name="Convert from inches to meters",
        default=False,
        description="Scale by 2.54/100",
    )
    scale = FloatProperty(
        name="Scale",
        default=1.0,
        description="Scale the positions by this value",
        min=0.0000001, max=1000000.0,
        soft_min=0.001, soft_max=100.0,
    )
    auto_scale = BoolProperty(
        name="Adjust scale automatically",
        default=False,
        description="Guess correct scale factor",
    )
    auto_magnitude = BoolProperty(
        name="Adjust scale magnitude",
        default=True,
        description="Automatically adjust scale magnitude",
    )
    create_armature = BoolProperty(
        name="Create an armature",
        default=True,
        description="Import the markers as bones instead of empties",
    )
    size = FloatProperty(
        name="Empty or bone size",
        default=.03,
        description="The size of each empty or bone",
        min=0.0001, max=1000000.0,
        soft_min=0.001, soft_max=100.0,
    )
    x_ray = BoolProperty(
        name="Use X-Ray",
        default=True,
        description="Show the empties or armature over other objects",
    )
    frame_skip = IntProperty(
        name="Fps divisor",
        default=1,
        description="Frame supersampling factor",
        min=1,
    )
    use_frame_no = BoolProperty(
        name="Use frame numbers",
        default=False,
        description="Offset start of animation according to the source",
    )
    show_names = BoolProperty(
        name="Show Names", default=False,
        description="Show the markers' name",
    )
    prefix = StringProperty(
        name="Name Prefix", maxlen=32,
        description="Prefix object names with this",
    )
    use_existing = BoolProperty(
        name="Use existing empties or armature",
        default=False,
        description="Use previously created homonymous empties or bones",
    )
    confidence = FloatProperty(
        name="Minimum Confidence Level", default=0,
        description="Only consider markers with at least "
                    "this confidence level",
        min=-1., max=1000000.0,
        soft_min=-1., soft_max=100.0,
    )

    filter_glob = StringProperty(default="*.c3d;*.csv", options={'HIDDEN'})

    def find_height(self, ms):
        """
            Heuristic to find the height of the subject in the markerset
            (only works for standing poses and you must have correct data
            on the first frame)
        """
        zmin, zmax = None, None
        hidx = 1 if self.properties.Y_up else 2
        for ml in ms.markerLabels:
            # check if LTOE is a substring of this marker label
            if 'LTOE' in ml:
                # substitute the substring to get the head marker
                hd = ml.replace('LTOE', 'LFHD')
                if hd not in ms.markerLabels:
                    break
                pmin_idx = ms.markerLabels.index(ml)
                pmax_idx = ms.markerLabels.index(hd)
                zmin = ms.frames[0][pmin_idx].position[hidx]
                zmax = ms.frames[0][pmax_idx].position[hidx]
                break
        if zmin is None:  # could not find named markers, get extremes
            allz = [m.position[hidx] for m in ms.frames[0]]
            zmin, zmax = min(allz), max(allz)
        return abs(zmax - zmin) or 1

    def adjust_scale_magnitude(self, height, scale):
        mag = math.log10(height * scale)
        return scale * math.pow(10, -int(mag))

    def adjust_scale(self, height, scale):
        """
            Try to find the correct scale for some common configurations
            found in CMU's c3d files.
        """
        factor = height * scale / 1.75  # normalize
        if factor < 0.5:
            scale /= 10.0
            factor *= 10.0
        cmu_factors = [(1.0, 1.0), (1.1, 1.45), (1.6, 1.6), (2.54, 2.54)]
        sqerr, fix = min(
            ((cf[0] - factor) ** 2.0, 1.0 / cf[1])
            for cf in cmu_factors
        )
        return scale * fix

    def create_empties(self, ms):
        """
            Create the empties and get their collision-free names
        """
        unames = {}
        use_existing = self.properties.use_existing
        s = self.properties.size
        empty_size = (s, s, s)
        for ml in ms.markerLabels:
            name = self.properties.prefix + ml
            if use_existing and name in bpy.context.scene.objects:
                o = bpy.context.scene.objects[name]
            else:
                bpy.ops.object.add()
                o = bpy.context.active_object
                o.name = name
            unames[name] = o.name
            bpy.ops.transform.resize(value=empty_size)
            o.show_name = self.properties.show_names
            o.show_x_ray = self.properties.x_ray
        for name in unames.values():
            bpy.context.scene.objects[name].select = True
        return unames

    def create_armature_obj(self, ms, scale):
        """
            Create or use existing armature, return a bone dict,
            leave the armature in POSE mode
        """
        head_dir = Vector((0, 0, self.properties.size))
        ao = bpy.context.active_object
        # when using an existing armature we restrict importing
        # the markers only for existing bones
        if not self.properties.use_existing or not ao or ao.type != 'ARMATURE':
            bpy.ops.object.add(type='ARMATURE', enter_editmode=True)
            arm = bpy.context.active_object
            arm.name = os.path.basename(self.properties.filepath)
            arm.data.show_names = self.properties.show_names
            arm.show_x_ray = self.properties.x_ray
            for idx, ml in enumerate(ms.markerLabels):
                name = self.properties.prefix + ml
                bpy.ops.armature.select_all(action='DESELECT')
                bpy.ops.armature.bone_primitive_add(name=name)
                pos = Vector(ms.frames[0][idx].position) * scale
                if self.properties.Y_up:
                    pos = Vector((pos[0], -pos[2], pos[1]))
                b = arm.data.edit_bones[name]
                b.head = pos + head_dir
                b.tail = pos
        bpy.ops.object.mode_set(mode='POSE')
        bpy.ops.pose.select_all(action='SELECT')

    def execute(self, context):
        ms = import_c3d.read(self.properties.filepath, onlyHeader=True)
        ms.readNextFrameData()

        # determine the final scale
        height = self.find_height(ms)
        scale = 1.0 if not self.properties.from_inches else 0.0254
        scale *= ms.scale
        if self.properties.auto_magnitude:
            scale = self.adjust_scale_magnitude(height, scale)
        if self.properties.auto_scale:
            scale = self.adjust_scale(height, scale)
        scale *= self.properties.scale

        if bpy.context.mode != 'OBJECT':
            bpy.ops.object.mode_set(mode='OBJECT')
        if self.properties.create_armature:
            self.create_armature_obj(ms, scale)
        else:
            unames = self.create_empties(ms)

        # start animating the empties
        C3DAnimateCloud.markerset = ms
        C3DAnimateCloud.is_armature = self.properties.create_armature
        if not C3DAnimateCloud.is_armature:
            C3DAnimateCloud.unames = unames
        C3DAnimateCloud.scale = scale
        C3DAnimateCloud.Y_up = self.properties.Y_up
        C3DAnimateCloud.fskip = self.properties.frame_skip
        C3DAnimateCloud.prefix = self.properties.prefix
        C3DAnimateCloud.use_frame_no = self.properties.use_frame_no
        C3DAnimateCloud.confidence = self.properties.confidence
        C3DAnimateCloud.curframe = ms.startFrame
        bpy.ops.import_anim.c3danim()
        return {'FINISHED'}

    def invoke(self, context, event):
        wm = context.window_manager
        wm.fileselect_add(self)
        return {'RUNNING_MODAL'}


def menu_func(self, context):
    self.layout.operator(C3DImporter.bl_idname,
                         text="Graphics Lab Motion Capture (.c3d)")


def register():
    bpy.utils.register_module(__name__)
    bpy.types.INFO_MT_file_import.append(menu_func)


def unregister():
    bpy.utils.unregister_module(__name__)
    bpy.types.INFO_MT_file_import.remove(menu_func)


if __name__ == "__main__":
    register()



import bpy, re
from ..widgets import create_hand_widget, create_gear_widget
from .ui             import create_script
from .limb_utils     import *
from mathutils       import Vector
from ...utils       import copy_bone, flip_bone, put_bone, create_cube_widget
from ...utils       import strip_org, strip_mch, make_deformer_name, create_widget
from ...utils       import create_circle_widget, create_sphere_widget, create_line_widget
from ...utils       import MetarigError, make_mechanism_name, org
from ...utils       import create_limb_widget, connected_children_names
from ...utils       import align_bone_y_axis, align_bone_x_axis, align_bone_z_axis
from rna_prop_ui import rna_idprop_ui_prop_get
from ..widgets import create_ikarrow_widget
from math import trunc, pi

extra_script = """
controls = [%s]
ctrl    = '%s'

if is_selected( controls ):
    layout.prop( pose_bones[ ctrl ], '["%s"]')
    if '%s' in pose_bones[ctrl].keys():
        layout.prop( pose_bones[ ctrl ], '["%s"]', slider = True )
    if '%s' in pose_bones[ctrl].keys():
        layout.prop( pose_bones[ ctrl ], '["%s"]', slider = True )
"""

IMPLEMENTATION = True   # Include and set True if Rig is just an implementation for a wrapper class
                        # add_parameters and parameters_ui are unused for implementation classes


class Rig:

    def __init__(self, obj, bone_name, params):
        """ Initialize arm rig and key rig properties """
        self.obj = obj
        self.params = params

        self.org_bones = list(
            [bone_name] + connected_children_names(obj, bone_name)
            )[:3]  # The basic limb is the first 3 bones

        self.segments = params.segments
        self.bbones = params.bbones
        self.limb_type = params.limb_type
        self.rot_axis = params.rotation_axis
        self.auto_align_extremity = params.auto_align_extremity

        # Assign values to tweak/FK layers props if opted by user
        if params.tweak_extra_layers:
            self.tweak_layers = list(params.tweak_layers)
        else:
            self.tweak_layers = None

        if params.fk_extra_layers:
            self.fk_layers = list(params.fk_layers)
        else:
            self.fk_layers = None

    def orient_org_bones(self):

        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones

        thigh = self.org_bones[0]
        org_bones = list(
            [thigh] + connected_children_names(self.obj, thigh)
        )  # All the provided orgs

        org_uarm = eb[org_bones[0]]
        org_farm = eb[org_bones[1]]
        org_hand = eb[org_bones[2]]

        if self.rot_axis != 'automatic':
            if self.auto_align_extremity:
                z_ground_projection = Vector((org_hand.z_axis[0], org_hand.z_axis[1], 0))
                align_bone_z_axis(self.obj, org_hand.name, z_ground_projection.normalized())
            return

        # Orient uarm farm bones
        chain_y_axis = org_uarm.y_axis + org_farm.y_axis
        chain_rot_axis = org_uarm.y_axis.cross(chain_y_axis).normalized()  # ik-plane normal axis (rotation)

        align_bone_x_axis(self.obj, org_uarm.name, chain_rot_axis)
        align_bone_x_axis(self.obj, org_farm.name, chain_rot_axis)

        # Orient hand
        align_bone_x_axis(self.obj, org_hand.name, chain_rot_axis)

    def create_parent(self):

        org_bones = self.org_bones

        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones

        name = get_bone_name( strip_org( org_bones[0] ), 'mch', 'parent' )

        mch = copy_bone( self.obj, org_bones[0], name )
        orient_bone( self, eb[mch], 'y' )
        eb[ mch ].length = eb[ org_bones[0] ].length / 4

        eb[ mch ].parent = eb[ org_bones[0] ].parent

        eb[ mch ].roll = 0.0

        # Add non-MCH main limb control
        name = get_bone_name(strip_org(org_bones[0]), 'ctrl', 'parent')
        main_parent = copy_bone(self.obj, org_bones[0], name)
        eb[main_parent].length = eb[org_bones[0]].length / 4
        eb[main_parent].parent = None
        eb[main_parent].roll = 0.0

        # Constraints
        make_constraint( self, mch, {
            'constraint'  : 'COPY_ROTATION',
            'subtarget'   : 'root'
        })

        make_constraint( self, mch, {
            'constraint'  : 'COPY_SCALE',
            'subtarget'   : 'root'
        })

        # Limb Follow Driver
        pb = self.obj.pose.bones

        name = 'FK_limb_follow'

        # pb[ mch ][ name ] = 0.0
        # prop = rna_idprop_ui_prop_get( pb[ mch ], name, create = True )
        pb[main_parent][name] = 0.0
        prop = rna_idprop_ui_prop_get(pb[main_parent], name, create=True)

        prop["min"] = 0.0
        prop["max"] = 1.0
        prop["soft_min"] = 0.0
        prop["soft_max"] = 1.0
        prop["description"] = name

        drv = pb[mch].constraints[0].driver_add("influence").driver

        drv.type = 'AVERAGE'
        var = drv.variables.new()
        var.name = name
        var.type = "SINGLE_PROP"
        var.targets[0].id = self.obj
        var.targets[0].data_path = pb[main_parent].path_from_id() + \
                                   '[' + '"' + name + '"' + ']'

        size = pb[main_parent].bone.y_axis.length * 10
        create_gear_widget(self.obj, main_parent, size=size, bone_transform_name=None)

        return [mch, main_parent]

    def create_tweak(self):
        org_bones = self.org_bones

        bpy.ops.object.mode_set(mode ='EDIT')
        eb = self.obj.data.edit_bones

        tweaks         = {}
        tweaks['ctrl'] = []
        tweaks['mch' ] = []

        # Create and parent mch and ctrl tweaks
        for i,org in enumerate(org_bones):
            if i < len(org_bones) - 1:
                # Create segments if specified
                for j in range( self.segments ):
                    # MCH
                    name = get_bone_name( strip_org(org), 'mch', 'tweak' )
                    mch = copy_bone( self.obj, org, name )

                    # CTRL
                    name = get_bone_name( strip_org(org), 'ctrl', 'tweak' )
                    ctrl = copy_bone( self.obj, org, name )

                    eb[ mch  ].length /= self.segments
                    eb[ ctrl ].length /= self.segments

                    # If we have more than one segments, place the head of the
                    # 2nd and onwards at the correct position
                    if j > 0:
                        put_bone(self.obj, mch,  eb[ tweaks['mch' ][-1] ].tail)
                        put_bone(self.obj, ctrl, eb[ tweaks['ctrl'][-1] ].tail)

                    tweaks['ctrl'] += [ ctrl ]
                    tweaks['mch' ] += [ mch  ]

                    # Parenting the tweak ctrls to mchs
                    eb[ mch  ].parent = eb[ org ]
                    eb[ ctrl ].parent = eb[ mch ]

            else: # Last limb bone - is not subdivided
                name = get_bone_name( strip_org(org), 'mch', 'tweak' )
                mch = copy_bone( self.obj, org_bones[i-1], name )
                eb[ mch ].length = eb[org].length / 4
                put_bone(
                    self.obj,
                    mch,
                    eb[org_bones[i-1]].tail
                )

                ctrl = get_bone_name( strip_org(org), 'ctrl', 'tweak' )
                ctrl = copy_bone( self.obj, org, ctrl )
                eb[ ctrl ].length = eb[org].length / 2

                tweaks['mch']  += [ mch  ]
                tweaks['ctrl'] += [ ctrl ]

                # Parenting the tweak ctrls to mchs
                eb[ mch  ].parent = eb[ org ]
                eb[ ctrl ].parent = eb[ mch ]

        # Scale to reduce widget size and maintain conventions!
        for mch, ctrl in zip( tweaks['mch'], tweaks['ctrl'] ):
            eb[ mch  ].length /= 4
            eb[ ctrl ].length /= 2

        # Contraints

        for i,b in enumerate( tweaks['mch'] ):
            first  = 0
            middle = trunc( len( tweaks['mch'] ) / 2 )
            last   = len( tweaks['mch'] ) - 1

            if i == first or i == middle:
                make_constraint( self, b, {
                    'constraint'  : 'COPY_SCALE',
                    'subtarget'   : 'root'
                })
            elif i != last:
                targets       = []
                dt_target_idx = middle
                factor        = 0
                if i < middle:
                    targets = [first,middle]
                else:
                    targets       = [middle,last]
                    factor        = self.segments
                    dt_target_idx = last

                # Use copy transforms constraints to position each bone
                # exactly in the location respective to its index (between
                # the two edges)
                make_constraint( self, b, {
                    'constraint'  : 'COPY_TRANSFORMS',
                    'subtarget'   : tweaks['ctrl'][targets[0]]
                })
                make_constraint( self, b, {
                    'constraint'  : 'COPY_TRANSFORMS',
                    'subtarget'   : tweaks['ctrl'][targets[1]],
                    'influence'   : (i - factor) / self.segments
                })
                make_constraint( self, b, {
                    'constraint'  : 'DAMPED_TRACK',
                    'subtarget'   : tweaks['ctrl'][ dt_target_idx ],
                })

        # Ctrl bones Locks and Widgets
        pb = self.obj.pose.bones
        for t in tweaks['ctrl']:
            pb[t].lock_rotation = True, False, True
            pb[t].lock_scale    = False, True, False

            create_sphere_widget(self.obj, t, bone_transform_name=None)

            if self.tweak_layers:
                pb[t].bone.layers = self.tweak_layers

        return tweaks

    def create_def(self, tweaks):
        org_bones = self.org_bones

        bpy.ops.object.mode_set(mode ='EDIT')
        eb = self.obj.data.edit_bones

        def_bones = []
        for i, org in enumerate(org_bones):
            if i < len(org_bones) - 1:
                # Create segments if specified
                for j in range(self.segments):
                    name = get_bone_name(strip_org(org), 'def')
                    def_name = copy_bone(self.obj, org, name)

                    eb[def_name].length /= self.segments

                    # If we have more than one segments, place the 2nd and
                    # onwards on the tail of the previous bone
                    if j > 0:
                        put_bone(self.obj, def_name, eb[ def_bones[-1] ].tail)

                    def_bones += [def_name]
            else:
                name = get_bone_name(strip_org(org), 'def')
                def_name = copy_bone(self.obj, org, name)
                def_bones.append(def_name)

        # Parent deform bones
        for i,b in enumerate( def_bones ):
            if i > 0: # For all bones but the first (which has no parent)
                eb[b].parent      = eb[ def_bones[i-1] ] # to previous
                eb[b].use_connect = True

        # Constraint def to tweaks
        for d,t in zip(def_bones, tweaks):
            tidx = tweaks.index(t)

            make_constraint( self, d, {
                'constraint'  : 'COPY_TRANSFORMS',
                'subtarget'   : t
            })

            if tidx != len(tweaks) - 1:
                make_constraint( self, d, {
                    'constraint'  : 'DAMPED_TRACK',
                    'subtarget'   : tweaks[ tidx + 1 ],
                })

                make_constraint( self, d, {
                    'constraint'  : 'STRETCH_TO',
                    'subtarget'   : tweaks[ tidx + 1 ],
                })

        # Create bbone segments
        for bone in def_bones[:-1]:
            self.obj.data.bones[bone].bbone_segments = self.bbones

        self.obj.data.bones[ def_bones[0]  ].bbone_in = 0.0
        self.obj.data.bones[ def_bones[-2] ].bbone_out = 0.0
        self.obj.data.bones[ def_bones[-1] ].bbone_in = 0.0
        self.obj.data.bones[ def_bones[-1] ].bbone_out = 0.0


        # Rubber hose drivers
        pb = self.obj.pose.bones
        for i,t in enumerate( tweaks[1:-1] ):
            # Create custom property on tweak bone to control rubber hose
            name = 'rubber_tweak'

            if i == trunc( len( tweaks[1:-1] ) / 2 ):
                pb[t][name] = 0.0
            else:
                pb[t][name] = 1.0

            prop = rna_idprop_ui_prop_get( pb[t], name, create=True )

            prop["min"]         = 0.0
            prop["max"]         = 2.0
            prop["soft_min"]    = 0.0
            prop["soft_max"]    = 1.0
            prop["description"] = name

        for j,d in enumerate(def_bones[:-1]):
            drvs = {}
            if j != 0:
                tidx = j
                drvs[tidx] = self.obj.data.bones[d].driver_add("bbone_in").driver

            if j != len( def_bones[:-1] ) - 1:
                tidx = j + 1
                drvs[tidx] = self.obj.data.bones[d].driver_add("bbone_out").driver

            for d in drvs:
                drv = drvs[d]
                name = 'rubber_tweak'
                drv.type = 'AVERAGE'
                var = drv.variables.new()
                var.name = name
                var.type = "SINGLE_PROP"
                var.targets[0].id = self.obj
                var.targets[0].data_path = pb[tweaks[d]].path_from_id() + \
                                           '[' + '"' + name + '"' + ']'

        return def_bones

    def create_ik(self, parent):
        org_bones = self.org_bones

        bpy.ops.object.mode_set(mode ='EDIT')
        eb = self.obj.data.edit_bones

        ctrl = get_bone_name(org_bones[0], 'ctrl', 'ik')
        mch_ik = get_bone_name(org_bones[0], 'mch', 'ik')
        mch_target = get_bone_name(org_bones[0], 'mch', 'ik_target')

        for o, ik in zip( org_bones, [ ctrl, mch_ik, mch_target ] ):
            bone = copy_bone( self.obj, o, ik )

            if org_bones.index(o) == len( org_bones ) - 1:
                eb[ bone ].length /= 4

        # Create MCH Stretch
        mch_str = copy_bone(
            self.obj,
            org_bones[0],
            get_bone_name( org_bones[0], 'mch', 'ik_stretch' )
        )

        eb[ mch_str ].tail = eb[ org_bones[-1] ].head

        # Parenting
        eb[ctrl].parent = eb[parent]
        eb[mch_str].parent = eb[parent]
        eb[mch_ik].parent = eb[ctrl]

        # Make standard pole target bone
        pole_name = get_bone_name(org_bones[0], 'ctrl', 'ik_target')
        pole_target = copy_bone(self.obj, org_bones[0], pole_name)

        lo_vector = eb[org_bones[1]].tail - eb[org_bones[1]].head
        tot_vector = eb[org_bones[0]].head - eb[org_bones[1]].tail
        tot_vector.normalize()
        elbow_vector = lo_vector.dot(tot_vector)*tot_vector - lo_vector    # elbow_vec as rejection of lo on tot
        elbow_vector.normalize()
        elbow_vector *= (eb[org_bones[1]].tail - eb[org_bones[0]].head).length

        if self.rot_axis == 'x' or self.rot_axis == 'automatic':
            z_vector = eb[org_bones[0]].z_axis + eb[org_bones[1]].z_axis
            alfa = elbow_vector.angle(z_vector)
        elif self.rot_axis == 'z':
            x_vector = eb[org_bones[0]].x_axis + eb[org_bones[1]].x_axis
            alfa = elbow_vector.angle(x_vector)

        if alfa > pi/2:
            pole_angle = -pi/2
        else:
            pole_angle = pi/2

        if self.rot_axis == 'z':
            pole_angle = 0

        eb[pole_target].head = eb[org_bones[0]].tail + elbow_vector
        eb[pole_target].tail = eb[pole_target].head - elbow_vector/8
        eb[pole_target].roll = 0.0

        # Make visual pole
        vispole_name = 'VIS_' + get_bone_name(org_bones[0], 'ctrl', 'ik_pole')
        vispole = copy_bone(self.obj, org_bones[1], vispole_name)
        eb[vispole].tail = eb[vispole].head + Vector((0.0, 0.0, eb[org_bones[1]].length/10))
        eb[vispole].use_connect = False
        eb[vispole].hide_select = True
        eb[vispole].parent = None

        make_constraint(self, mch_ik, {
            'constraint': 'IK',
            'subtarget': mch_target,
            'chain_count': 2,
        })

        make_constraint(self, mch_ik, {     # 2_nd IK for pole targeted chain
            'constraint': 'IK',
            'subtarget': mch_target,
            'chain_count': 2,
        })

        # VIS pole constraints
        make_constraint(self, vispole, {
            'constraint': 'COPY_LOCATION',
            'name': 'copy_loc',
            'subtarget': org_bones[1],
        })

        pb = self.obj.pose.bones

        make_constraint(self, vispole, {
            'constraint': 'STRETCH_TO',
            'name': 'stretch_to',
            'subtarget': pole_target,
            'volume': 'NO_VOLUME',
            'rest_length': pb[vispole].length
        })

        pb[mch_ik].constraints[-1].pole_target = self.obj
        pb[mch_ik].constraints[-1].pole_subtarget = pole_target
        pb[mch_ik].constraints[-1].pole_angle = pole_angle

        pb[ mch_ik ].ik_stretch = 0.1
        pb[ ctrl   ].ik_stretch = 0.1

        # IK constraint Rotation locks
        if self.rot_axis == 'z':
            pb[mch_ik].lock_ik_x = True
            pb[mch_ik].lock_ik_y = True
        else:
            pb[mch_ik].lock_ik_y = True
            pb[mch_ik].lock_ik_z = True

        # Locks and Widget
        pb[ctrl].lock_rotation = True, False, True
        if self.rot_axis == 'x' or self.rot_axis == 'automatic':
            roll = 0
        else:
            roll = pi/2
        create_ikarrow_widget(self.obj, ctrl, bone_transform_name=None, roll=roll)
        create_sphere_widget(self.obj, pole_target, bone_transform_name=None)
        create_line_widget(self.obj, vispole)

        return {'ctrl': {'limb': ctrl, 'ik_target': pole_target},
                'mch_ik': mch_ik,
                'mch_target': mch_target,
                'mch_str': mch_str,
                'visuals': {'vispole': vispole}
        }

    def create_fk(self, parent):
        org_bones = self.org_bones.copy()

        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones

        ctrls = []

        for o in org_bones:
            bone = copy_bone(self.obj, o, get_bone_name( o, 'ctrl', 'fk'))
            ctrls.append(bone)

        # MCH
        mch = copy_bone(
            self.obj, org_bones[-1], get_bone_name(o, 'mch', 'fk')
        )

        eb[mch].length /= 4

        # Parenting
        eb[ctrls[0]].parent = eb[parent]
        eb[ctrls[1]].parent = eb[ctrls[0]]
        eb[ctrls[1]].use_connect = True
        eb[ctrls[2]].parent = eb[mch]
        eb[mch].parent = eb[ctrls[1]]
        eb[mch].use_connect = True

        # Constrain MCH's scale to root
        make_constraint(self, mch, {
            'constraint': 'COPY_SCALE',
            'subtarget': 'root'
        })

        # Locks and widgets
        pb = self.obj.pose.bones
        pb[ctrls[2]].lock_location = True, True, True

        create_limb_widget(self.obj, ctrls[0])
        create_limb_widget(self.obj, ctrls[1])

        create_circle_widget(self.obj, ctrls[2], radius=0.4, head_tail=0.0)

        for c in ctrls:
            if self.fk_layers:
                pb[c].bone.layers = self.fk_layers

        return {'ctrl': ctrls, 'mch': mch}

    def org_parenting_and_switch(self, org_bones, ik, fk, parent):
        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones
        # re-parent ORGs in a connected chain
        for i, o in enumerate(org_bones):
            if i > 0:
                eb[o].parent = eb[org_bones[i-1]]
                if i <= len(org_bones)-1:
                    eb[o].use_connect = True

        bpy.ops.object.mode_set(mode='OBJECT')
        pb = self.obj.pose.bones
        pb_parent = pb[parent]

        # Create ik/fk switch property
        pb_parent['IK_FK'] = 0.0
        prop = rna_idprop_ui_prop_get(pb_parent, 'IK_FK', create=True)
        prop["min"] = 0.0
        prop["max"] = 1.0
        prop["soft_min"] = 0.0
        prop["soft_max"] = 1.0
        prop["description"] = 'IK/FK Switch'

        # Constrain org to IK and FK bones
        iks = [ik['ctrl']['limb']]
        iks += [ik[k] for k in ['mch_ik', 'mch_target']]

        for o, i, f in zip(org_bones, iks, fk):
            make_constraint( self, o, {
                'constraint': 'COPY_TRANSFORMS',
                'subtarget': i
            })
            make_constraint(self, o, {
                'constraint': 'COPY_TRANSFORMS',
                'subtarget': f
            })

            # Add driver to relevant constraint
            drv = pb[o].constraints[-1].driver_add("influence").driver
            drv.type = 'AVERAGE'

            var = drv.variables.new()
            var.name = prop.name
            var.type = "SINGLE_PROP"
            var.targets[0].id = self.obj
            var.targets[0].data_path = \
                pb_parent.path_from_id() + '[' + '"' + prop.name + '"' + ']'

    def create_arm(self, bones):
        org_bones = self.org_bones

        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones

        pole_target = get_bone_name(org_bones[0], 'ctrl', 'ik_target')

        # Create IK arm control
        ctrl = get_bone_name(org_bones[2], 'ctrl', 'ik')
        ctrl = copy_bone(self.obj, org_bones[2], ctrl)

        # clear parent (so that rigify will parent to root)
        eb[ctrl].parent = None
        eb[ctrl].use_connect = False

        # Parent
        eb[ bones['ik']['mch_target'] ].parent      = eb[ ctrl ]
        eb[ bones['ik']['mch_target'] ].use_connect = False

        # MCH for ik control
        ctrl_socket = copy_bone(self.obj, org_bones[2], get_bone_name( org_bones[2], 'mch', 'ik_socket'))
        eb[ctrl_socket].tail = eb[ctrl_socket].head + 0.8*(eb[ctrl_socket].tail-eb[ctrl_socket].head)
        eb[ctrl_socket].parent = None
        eb[ctrl].parent = eb[ctrl_socket]

        # MCH for pole ik control
        ctrl_pole_socket = copy_bone(self.obj, org_bones[2], get_bone_name(org_bones[2], 'mch', 'pole_ik_socket'))
        eb[ctrl_pole_socket].tail = eb[ctrl_pole_socket].head + 0.8 * (eb[ctrl_pole_socket].tail - eb[ctrl_pole_socket].head)
        eb[ctrl_pole_socket].parent = None
        eb[pole_target].parent = eb[ctrl_pole_socket]

        ctrl_root = copy_bone(self.obj, org_bones[2], get_bone_name( org_bones[2], 'mch', 'ik_root'))
        eb[ctrl_root].tail = eb[ctrl_root].head + 0.7*(eb[ctrl_root].tail-eb[ctrl_root].head)
        eb[ctrl_root].use_connect = False
        eb[ctrl_root].parent = eb['root']

        if eb[org_bones[0]].parent:
            arm_parent = eb[org_bones[0]].parent
            ctrl_parent = copy_bone(self.obj, org_bones[2], get_bone_name( org_bones[2], 'mch', 'ik_parent'))
            eb[ctrl_parent].tail = eb[ctrl_parent].head + 0.6*(eb[ctrl_parent].tail-eb[ctrl_parent].head)
            eb[ctrl_parent].use_connect = False
            if eb[org_bones[0]].parent_recursive:
                eb[ctrl_parent].parent = eb[org_bones[0]].parent_recursive[-1]
            else:
                eb[ctrl_parent].parent = eb[org_bones[0]].parent
        else:
            arm_parent = None

        mch_name = get_bone_name(strip_org(org_bones[0]), 'mch', 'parent_socket')
        mch_main_parent = copy_bone(self.obj, org_bones[0], mch_name)
        eb[mch_main_parent].length = eb[org_bones[0]].length / 12
        eb[mch_main_parent].parent = eb[bones['parent']]
        eb[mch_main_parent].roll = 0.0
        eb[bones['main_parent']].parent = eb[mch_main_parent]

        # Set up constraints

        # Constrain ik ctrl to root / parent

        make_constraint( self, ctrl_socket, {
            'constraint'  : 'COPY_TRANSFORMS',
            'subtarget'   : ctrl_root,
        })

        make_constraint(self, ctrl_pole_socket, {
            'constraint': 'COPY_TRANSFORMS',
            'subtarget': ctrl_root,
        })

        if arm_parent:
            make_constraint( self, ctrl_socket, {
                'constraint'  : 'COPY_TRANSFORMS',
                'subtarget'   : ctrl_parent,
            })

            make_constraint(self, ctrl_pole_socket, {
                'constraint': 'COPY_TRANSFORMS',
                'subtarget': ctrl_parent,
            })

        # Constrain mch target bone to the ik control and mch stretch
        make_constraint( self, bones['ik']['mch_target'], {
            'constraint'  : 'COPY_LOCATION',
            'subtarget'   : bones['ik']['mch_str'],
            'head_tail'   : 1.0
        })

        # Constrain mch ik stretch bone to the ik control
        make_constraint(self, bones['ik']['mch_str'], {
            'constraint': 'DAMPED_TRACK',
            'subtarget': ctrl,
        })
        make_constraint(self, bones['ik']['mch_str'], {
            'constraint': 'STRETCH_TO',
            'subtarget': ctrl,
        })
        make_constraint(self, bones['ik']['mch_str'], {
            'constraint': 'LIMIT_SCALE',
            'use_min_y': True,
            'use_max_y': True,
            'max_y': 1.05,
            'owner_space': 'LOCAL'
        })
        make_constraint(self, mch_main_parent, {
            'constraint': 'COPY_ROTATION',
            'subtarget': org_bones[0]
        })

        pb = self.obj.pose.bones

        # Create ik/fk switch property
        pb_parent = pb[bones['main_parent']]
        pb_parent.lock_location = True, True, True
        pb_parent.lock_rotation = True, True, True
        pb_parent.lock_scale = True, True, True

        # Modify rotation mode for ik and tweak controls
        pb[bones['ik']['ctrl']['limb']].rotation_mode = 'ZXY'

        for b in bones['tweak']['ctrl']:
            pb[b].rotation_mode = 'ZXY'

        pb_parent['IK_Stretch'] = 1.0
        prop = rna_idprop_ui_prop_get(pb_parent, 'IK_Stretch', create=True)
        prop["min"] = 0.0
        prop["max"] = 1.0
        prop["soft_min"] = 0.0
        prop["soft_max"] = 1.0
        prop["description"] = 'IK Stretch'

        # Add driver to limit scale constraint influence
        b = bones['ik']['mch_str']
        drv = pb[b].constraints[-1].driver_add("influence").driver
        drv.type = 'SUM'

        var = drv.variables.new()
        var.name = prop.name
        var.type = "SINGLE_PROP"
        var.targets[0].id = self.obj
        var.targets[0].data_path = \
            pb_parent.path_from_id() + '[' + '"' + prop.name + '"' + ']'

        drv_modifier = self.obj.animation_data.drivers[-1].modifiers[0]

        drv_modifier.mode = 'POLYNOMIAL'
        drv_modifier.poly_order = 1
        drv_modifier.coefficients[0] = 1.0
        drv_modifier.coefficients[1] = -1.0

        # Create hand widget
        create_hand_widget(self.obj, ctrl, bone_transform_name=None)

        bones['ik']['ctrl']['terminal'] = [ctrl]
        if arm_parent:
            bones['ik']['mch_hand'] = [ctrl_socket, ctrl_pole_socket, ctrl_root, ctrl_parent]
        else:
            bones['ik']['mch_hand'] = [ctrl_socket, ctrl_pole_socket, ctrl_root]

        return bones

    def create_drivers(self, bones):

        bpy.ops.object.mode_set(mode='OBJECT')
        pb = self.obj.pose.bones

        ctrl = pb[bones['ik']['mch_hand'][0]]
        ctrl_pole = pb[bones['ik']['mch_hand'][1]]

        #owner = pb[bones['ik']['ctrl']['limb']]
        owner = pb[bones['main_parent']]

        props = ["IK_follow", "root/parent", "pole_vector", "pole_follow"]

        for prop in props:

            if prop == 'pole_vector':
                owner[prop] = False
                pole_prop = rna_idprop_ui_prop_get(owner, prop, create=True)
                pole_prop["min"] = False
                pole_prop["max"] = True
                pole_prop["description"] = prop
                mch_ik = pb[bones['ik']['mch_ik']]

                # ik target hide driver
                pole_target = pb[bones['ik']['ctrl']['ik_target']]
                drv = pole_target.bone.driver_add("hide").driver
                drv.type = 'AVERAGE'

                var = drv.variables.new()
                var.name = prop
                var.type = "SINGLE_PROP"
                var.targets[0].id = self.obj
                var.targets[0].data_path = \
                    owner.path_from_id() + '[' + '"' + prop + '"' + ']'

                drv_modifier = self.obj.animation_data.drivers[-1].modifiers[0]

                drv_modifier.mode = 'POLYNOMIAL'
                drv_modifier.poly_order = 1
                drv_modifier.coefficients[0] = 1.0
                drv_modifier.coefficients[1] = -1.0

                # vis-pole hide driver
                vispole = pb[bones['ik']['visuals']['vispole']]
                drv = vispole.bone.driver_add("hide").driver
                drv.type = 'AVERAGE'
                var = drv.variables.new()
                var.name = prop
                var.type = "SINGLE_PROP"
                var.targets[0].id = self.obj
                var.targets[0].data_path = \
                    owner.path_from_id() + '[' + '"' + prop + '"' + ']'

                drv_modifier = self.obj.animation_data.drivers[-1].modifiers[0]

                drv_modifier.mode = 'POLYNOMIAL'
                drv_modifier.poly_order = 1
                drv_modifier.coefficients[0] = 1.0
                drv_modifier.coefficients[1] = -1.0

                # arrow hide driver
                # pole_target = pb[bones['ik']['ctrl']['limb']]
                # drv = pole_target.bone.driver_add("hide").driver
                # drv.type = 'AVERAGE'
                #
                # var = drv.variables.new()
                # var.name = prop
                # var.type = "SINGLE_PROP"
                # var.targets[0].id = self.obj
                # var.targets[0].data_path = \
                #     owner.path_from_id() + '[' + '"' + prop + '"' + ']'
                #
                # drv_modifier = self.obj.animation_data.drivers[-1].modifiers[0]
                #
                # drv_modifier.mode = 'POLYNOMIAL'
                # drv_modifier.poly_order = 1
                # drv_modifier.coefficients[0] = 0.0
                # drv_modifier.coefficients[1] = 1.0

                for cns in mch_ik.constraints:
                    if 'IK' in cns.type:
                        drv = cns.driver_add("mute").driver
                        drv.type = 'AVERAGE'

                        var = drv.variables.new()
                        var.name = prop
                        var.type = "SINGLE_PROP"
                        var.targets[0].id = self.obj
                        var.targets[0].data_path = \
                            owner.path_from_id() + '[' + '"' + prop + '"' + ']'

                        drv_modifier = self.obj.animation_data.drivers[-1].modifiers[0]

                        drv_modifier.mode = 'POLYNOMIAL'
                        drv_modifier.poly_order = 1
                        if not cns.pole_subtarget:
                            drv_modifier.coefficients[0] = 0.0
                            drv_modifier.coefficients[1] = 1
                        else:
                            drv_modifier.coefficients[0] = 1.0
                            drv_modifier.coefficients[1] = -1.0

            elif prop == 'IK_follow':

                owner[prop] = True
                rna_prop = rna_idprop_ui_prop_get(owner, prop, create=True)
                rna_prop["min"] = False
                rna_prop["max"] = True
                rna_prop["description"] = prop

                drv = ctrl.constraints[0].driver_add("mute").driver
                drv.type = 'AVERAGE'

                var = drv.variables.new()
                var.name = prop
                var.type = "SINGLE_PROP"
                var.targets[0].id = self.obj
                var.targets[0].data_path = \
                    owner.path_from_id() + '[' + '"' + prop + '"' + ']'

                drv_modifier = self.obj.animation_data.drivers[-1].modifiers[0]

                drv_modifier.mode = 'POLYNOMIAL'
                drv_modifier.poly_order = 1
                drv_modifier.coefficients[0] = 1.0
                drv_modifier.coefficients[1] = -1.0

                if len(ctrl.constraints) > 1:
                    drv = ctrl.constraints[1].driver_add("mute").driver
                    drv.type = 'AVERAGE'

                    var = drv.variables.new()
                    var.name = prop
                    var.type = "SINGLE_PROP"
                    var.targets[0].id = self.obj
                    var.targets[0].data_path = \
                        owner.path_from_id() + '[' + '"' + prop + '"' + ']'

                    drv_modifier = self.obj.animation_data.drivers[-1].modifiers[0]

                    drv_modifier.mode = 'POLYNOMIAL'
                    drv_modifier.poly_order = 1
                    drv_modifier.coefficients[0] = 1.0
                    drv_modifier.coefficients[1] = -1.0

                drv = ctrl_pole.constraints[0].driver_add("mute").driver
                drv.type = 'AVERAGE'

                var = drv.variables.new()
                var.name = prop
                var.type = "SINGLE_PROP"
                var.targets[0].id = self.obj
                var.targets[0].data_path = \
                    owner.path_from_id() + '[' + '"' + prop + '"' + ']'

                drv_modifier = self.obj.animation_data.drivers[-1].modifiers[0]

                drv_modifier.mode = 'POLYNOMIAL'
                drv_modifier.poly_order = 1
                drv_modifier.coefficients[0] = 1.0
                drv_modifier.coefficients[1] = -1.0

                if len(ctrl_pole.constraints) > 1:
                    drv = ctrl_pole.constraints[1].driver_add("mute").driver
                    drv.type = 'AVERAGE'

                    var = drv.variables.new()
                    var.name = prop
                    var.type = "SINGLE_PROP"
                    var.targets[0].id = self.obj
                    var.targets[0].data_path = \
                        owner.path_from_id() + '[' + '"' + prop + '"' + ']'

                    drv_modifier = self.obj.animation_data.drivers[-1].modifiers[0]

                    drv_modifier.mode = 'POLYNOMIAL'
                    drv_modifier.poly_order = 1
                    drv_modifier.coefficients[0] = 1.0
                    drv_modifier.coefficients[1] = -1.0

            elif prop == 'root/parent':
                if len(ctrl.constraints) > 1:
                    owner[prop] = 0.0
                    rna_prop = rna_idprop_ui_prop_get(owner, prop, create=True)
                    rna_prop["min"] = 0.0
                    rna_prop["max"] = 1.0
                    rna_prop["soft_min"] = 0.0
                    rna_prop["soft_max"] = 1.0
                    rna_prop["description"] = prop

                    drv = ctrl.constraints[1].driver_add("influence").driver
                    drv.type = 'AVERAGE'

                    var = drv.variables.new()
                    var.name = prop
                    var.type = "SINGLE_PROP"
                    var.targets[0].id = self.obj
                    var.targets[0].data_path = \
                        owner.path_from_id() + '[' + '"' + prop + '"' + ']'

            elif prop == 'pole_follow':
                if len(ctrl_pole.constraints) > 1:
                    owner[prop] = 0.0
                    rna_prop = rna_idprop_ui_prop_get(owner, prop, create=True)
                    rna_prop["min"] = 0.0
                    rna_prop["max"] = 1.0
                    rna_prop["soft_min"] = 0.0
                    rna_prop["soft_max"] = 1.0
                    rna_prop["description"] = prop

                    drv = ctrl_pole.constraints[1].driver_add("influence").driver
                    drv.type = 'AVERAGE'

                    var = drv.variables.new()
                    var.name = prop
                    var.type = "SINGLE_PROP"
                    var.targets[0].id = self.obj
                    var.targets[0].data_path = \
                        owner.path_from_id() + '[' + '"' + prop + '"' + ']'

    @staticmethod
    def get_future_names(bones):

        if len(bones) != 3:
            return

        names = dict()

        uarm = strip_mch(strip_org(bones[0].name))
        farm = strip_mch(strip_org(bones[1].name))
        hand = strip_mch(strip_org(bones[2].name))

        suffix=''
        if uarm[-2:] == '.L' or uarm[-2:] == '.R':
            suffix = uarm[-2:]
            uarm = uarm.rstrip(suffix)
            farm = farm.rstrip(suffix)
            hand = hand.rstrip(suffix)

        # the following is declared in rig_ui
        # controls = ['upper_arm_ik.L', 'upper_arm_fk.L', 'forearm_fk.L', 'hand_fk.L', 'hand_ik.L', 'MCH-hand_fk.L',
        #             'upper_arm_parent.L']
        # tweaks = ['upper_arm_tweak.L.001', 'forearm_tweak.L', 'forearm_tweak.L.001']
        # ik_ctrl = ['hand_ik.L', 'MCH-upper_arm_ik.L', 'MCH-upper_arm_ik_target.L']
        # fk_ctrl = 'upper_arm_fk.L'
        # parent = 'upper_arm_parent.L'
        # hand_fk = 'hand_fk.L'
        # pole = 'upper_arm_ik_target.L'

        names['controls'] = [uarm + '_ik', uarm + '_fk', farm + '_fk', hand + '_fk', hand + '_ik',
                             make_mechanism_name(hand + '_fk'), uarm + '_parent']
        names['ik_ctrl'] = [hand + '_ik', make_mechanism_name(uarm) + '_ik', make_mechanism_name(uarm) + '_ik_target']
        names['fk_ctrl'] = uarm + '_fk' + suffix
        names['parent'] = uarm + '_parent' + suffix
        names['hand_fk'] = hand + '_fk' + suffix
        names['pole'] = uarm + '_ik_target' + suffix
        names['limb_type'] = 'arm'

        if suffix:
            for i, name in enumerate(names['controls']):
                names['controls'][i] = name + suffix
            for i, name in enumerate(names['ik_ctrl']):
                names['ik_ctrl'][i] = name + suffix

        return names

    def generate(self):
        bpy.ops.object.mode_set(mode='EDIT')
        eb = self.obj.data.edit_bones

        # Adjust org-bones rotation
        self.orient_org_bones()

        # Clear parents for org bones
        for bone in self.org_bones[1:]:
            eb[bone].use_connect = False
            eb[bone].parent = None

        bones = {}

        # Create mch limb parent
        mch_parent, main_parent = self.create_parent()
        bones['parent'] = mch_parent
        bones['main_parent'] = main_parent
        bones['tweak'] = self.create_tweak()
        bones['def'] = self.create_def(bones['tweak']['ctrl'])
        bones['ik'] = self.create_ik(bones['parent'])
        bones['fk'] = self.create_fk(bones['parent'])

        self.org_parenting_and_switch(self.org_bones, bones['ik'], bones['fk']['ctrl'], bones['main_parent'])

        bones = self.create_arm(bones)
        self.create_drivers(bones)

        controls = [bones['ik']['ctrl']['limb'], bones['ik']['ctrl']['terminal'][0]]

        controls.append(bones['main_parent'])

        # Create UI
        controls_string = ", ".join(["'" + x + "'" for x in controls])

        script = create_script(bones, 'arm')
        script += extra_script % (controls_string, bones['main_parent'], 'IK_follow',
                                  'pole_follow', 'pole_follow', 'root/parent', 'root/parent')

        return [script]


def add_parameters(params):
    """ Add the parameters of this rig type to the
        RigifyParameters PropertyGroup
    """

    items = [
        ('x', 'X manual', ''),
        ('z', 'Z manual', ''),
        ('automatic', 'Automatic', '')
    ]

    params.rotation_axis = bpy.props.EnumProperty(
        items   = items,
        name    = "Rotation Axis",
        default = 'automatic'
    )

    params.auto_align_extremity = bpy.BoolProperty(
        name='auto_align_extremity',
        default=False,
        description="Auto Align Extremity Bone"
    )

    params.segments = bpy.props.IntProperty(
        name        = 'limb segments',
        default     = 2,
        min         = 1,
        description = 'Number of segments'
    )

    params.bbones = bpy.props.IntProperty(
        name        = 'bbone segments',
        default     = 10,
        min         = 1,
        description = 'Number of segments'
    )

    # Setting up extra layers for the FK and tweak
    params.tweak_extra_layers = bpy.props.BoolProperty(
        name        = "tweak_extra_layers",
        default     = True,
        description = ""
        )

    params.tweak_layers = bpy.props.BoolVectorProperty(
        size        = 32,
        description = "Layers for the tweak controls to be on",
        default     = tuple( [ i == 1 for i in range(0, 32) ] )
        )

    # Setting up extra layers for the FK and tweak
    params.fk_extra_layers = bpy.props.BoolProperty(
        name        = "fk_extra_layers",
        default     = True,
        description = ""
        )

    params.fk_layers = bpy.props.BoolVectorProperty(
        size        = 32,
        description = "Layers for the FK controls to be on",
        default     = tuple( [ i == 1 for i in range(0, 32) ] )
        )


def parameters_ui(layout, params):
    """ Create the ui for the rig parameters."""

    r = layout.row()
    r.prop(params, "rotation_axis")

    if 'auto' not in params.rotation_axis.lower():
        r = layout.row()
        text = "Auto align Hand"
        r.prop(params, "auto_align_extremity", text=text)

    r = layout.row()
    r.prop(params, "segments")

    r = layout.row()
    r.prop(params, "bbones")

    bone_layers = bpy.context.active_pose_bone.bone.layers[:]

    for layer in ['fk', 'tweak']:
        r = layout.row()
        r.prop(params, layer + "_extra_layers")
        r.active = params.tweak_extra_layers

        col = r.column(align=True)
        row = col.row(align=True)

        for i in range(8):
            icon = "NONE"
            if bone_layers[i]:
                icon = "LAYER_ACTIVE"
            row.prop(params, layer + "_layers", index=i, toggle=True, text="", icon=icon)

        row = col.row(align=True)

        for i in range(16, 24):
            icon = "NONE"
            if bone_layers[i]:
                icon = "LAYER_ACTIVE"
            row.prop(params, layer + "_layers", index=i, toggle=True, text="", icon=icon)

        col = r.column(align=True)
        row = col.row(align=True)

        for i in range(8, 16):
            icon = "NONE"
            if bone_layers[i]:
                icon = "LAYER_ACTIVE"
            row.prop(params, layer + "_layers", index=i, toggle=True, text="", icon=icon)

        row = col.row(align=True)

        for i in range(24, 32):
            icon = "NONE"
            if bone_layers[i]:
                icon = "LAYER_ACTIVE"
            row.prop(params, layer + "_layers", index=i, toggle=True, text="", icon=icon)


def create_sample(obj):
    # generated by rigify.utils.write_metarig
    bpy.ops.object.mode_set(mode='EDIT')
    arm = obj.data

    bones = {}

    bone = arm.edit_bones.new('upper_arm.L')
    bone.head[:] = 0.1953, 0.0267, 1.5846
    bone.tail[:] = 0.4424, 0.0885, 1.4491
    bone.roll = 2.0724
    bone.use_connect = False
    bones['upper_arm.L'] = bone.name
    bone = arm.edit_bones.new('forearm.L')
    bone.head[:] = 0.4424, 0.0885, 1.4491
    bone.tail[:] = 0.6594, 0.0492, 1.3061
    bone.roll = 2.1535
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['upper_arm.L']]
    bones['forearm.L'] = bone.name
    bone = arm.edit_bones.new('hand.L')
    bone.head[:] = 0.6594, 0.0492, 1.3061
    bone.tail[:] = 0.7234, 0.0412, 1.2585
    bone.roll = 2.2103
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['forearm.L']]
    bones['hand.L'] = bone.name
    bone = arm.edit_bones.new('palm.01.L')
    bone.head[:] = 0.6921, 0.0224, 1.2882
    bone.tail[:] = 0.7464, 0.0051, 1.2482
    bone.roll = -2.4928
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['hand.L']]
    bones['palm.01.L'] = bone.name
    bone = arm.edit_bones.new('palm.02.L')
    bone.head[:] = 0.6970, 0.0389, 1.2877
    bone.tail[:] = 0.7518, 0.0277, 1.2487
    bone.roll = -2.5274
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['hand.L']]
    bones['palm.02.L'] = bone.name
    bone = arm.edit_bones.new('palm.03.L')
    bone.head[:] = 0.6963, 0.0545, 1.2874
    bone.tail[:] = 0.7540, 0.0521, 1.2482
    bone.roll = -2.5843
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['hand.L']]
    bones['palm.03.L'] = bone.name
    bone = arm.edit_bones.new('palm.04.L')
    bone.head[:] = 0.6929, 0.0696, 1.2871
    bone.tail[:] = 0.7528, 0.0763, 1.2428
    bone.roll = -2.5155
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['hand.L']]
    bones['palm.04.L'] = bone.name
    bone = arm.edit_bones.new('f_index.01.L')
    bone.head[:] = 0.7464, 0.0051, 1.2482
    bone.tail[:] = 0.7718, 0.0013, 1.2112
    bone.roll = -2.0315
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['palm.01.L']]
    bones['f_index.01.L'] = bone.name
    bone = arm.edit_bones.new('thumb.01.L')
    bone.head[:] = 0.6705, 0.0214, 1.2738
    bone.tail[:] = 0.6857, 0.0015, 1.2404
    bone.roll = -0.1587
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['palm.01.L']]
    bones['thumb.01.L'] = bone.name
    bone = arm.edit_bones.new('f_middle.01.L')
    bone.head[:] = 0.7518, 0.0277, 1.2487
    bone.tail[:] = 0.7762, 0.0234, 1.2058
    bone.roll = -2.0067
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['palm.02.L']]
    bones['f_middle.01.L'] = bone.name
    bone = arm.edit_bones.new('f_ring.01.L')
    bone.head[:] = 0.7540, 0.0521, 1.2482
    bone.tail[:] = 0.7715, 0.0499, 1.2070
    bone.roll = -2.0082
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['palm.03.L']]
    bones['f_ring.01.L'] = bone.name
    bone = arm.edit_bones.new('f_pinky.01.L')
    bone.head[:] = 0.7528, 0.0763, 1.2428
    bone.tail[:] = 0.7589, 0.0765, 1.2156
    bone.roll = -1.9749
    bone.use_connect = False
    bone.parent = arm.edit_bones[bones['palm.04.L']]
    bones['f_pinky.01.L'] = bone.name
    bone = arm.edit_bones.new('f_index.02.L')
    bone.head[:] = 0.7718, 0.0013, 1.2112
    bone.tail[:] = 0.7840, -0.0003, 1.1858
    bone.roll = -1.8799
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['f_index.01.L']]
    bones['f_index.02.L'] = bone.name
    bone = arm.edit_bones.new('thumb.02.L')
    bone.head[:] = 0.6857, 0.0015, 1.2404
    bone.tail[:] = 0.7056, -0.0057, 1.2145
    bone.roll = -0.4798
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['thumb.01.L']]
    bones['thumb.02.L'] = bone.name
    bone = arm.edit_bones.new('f_middle.02.L')
    bone.head[:] = 0.7762, 0.0234, 1.2058
    bone.tail[:] = 0.7851, 0.0218, 1.1749
    bone.roll = -1.8283
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['f_middle.01.L']]
    bones['f_middle.02.L'] = bone.name
    bone = arm.edit_bones.new('f_ring.02.L')
    bone.head[:] = 0.7715, 0.0499, 1.2070
    bone.tail[:] = 0.7794, 0.0494, 1.1762
    bone.roll = -1.8946
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['f_ring.01.L']]
    bones['f_ring.02.L'] = bone.name
    bone = arm.edit_bones.new('f_pinky.02.L')
    bone.head[:] = 0.7589, 0.0765, 1.2156
    bone.tail[:] = 0.7618, 0.0770, 1.1932
    bone.roll = -1.9059
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['f_pinky.01.L']]
    bones['f_pinky.02.L'] = bone.name
    bone = arm.edit_bones.new('f_index.03.L')
    bone.head[:] = 0.7840, -0.0003, 1.1858
    bone.tail[:] = 0.7892, 0.0006, 1.1636
    bone.roll = -1.6760
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['f_index.02.L']]
    bones['f_index.03.L'] = bone.name
    bone = arm.edit_bones.new('thumb.03.L')
    bone.head[:] = 0.7056, -0.0057, 1.2145
    bone.tail[:] = 0.7194, -0.0098, 1.1995
    bone.roll = -0.5826
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['thumb.02.L']]
    bones['thumb.03.L'] = bone.name
    bone = arm.edit_bones.new('f_middle.03.L')
    bone.head[:] = 0.7851, 0.0218, 1.1749
    bone.tail[:] = 0.7888, 0.0216, 1.1525
    bone.roll = -1.7483
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['f_middle.02.L']]
    bones['f_middle.03.L'] = bone.name
    bone = arm.edit_bones.new('f_ring.03.L')
    bone.head[:] = 0.7794, 0.0494, 1.1762
    bone.tail[:] = 0.7781, 0.0498, 1.1577
    bone.roll = -1.6582
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['f_ring.02.L']]
    bones['f_ring.03.L'] = bone.name
    bone = arm.edit_bones.new('f_pinky.03.L')
    bone.head[:] = 0.7618, 0.0770, 1.1932
    bone.tail[:] = 0.7611, 0.0772, 1.1782
    bone.roll = -1.7639
    bone.use_connect = True
    bone.parent = arm.edit_bones[bones['f_pinky.02.L']]
    bones['f_pinky.03.L'] = bone.name

    bpy.ops.object.mode_set(mode='OBJECT')
    pbone = obj.pose.bones[bones['upper_arm.L']]
    pbone.rigify_type = 'limbs.super_limb'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    try:
        pbone.rigify_parameters.separate_ik_layers = True
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.ik_layers = [False, False, False, False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.separate_hose_layers = True
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.hose_layers = [False, False, False, False, False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.tweak_layers = [False, False, False, False, False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.fk_layers = [False, False, False, False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['forearm.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['hand.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['palm.01.L']]
    pbone.rigify_type = 'limbs.super_palm'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'YXZ'
    pbone = obj.pose.bones[bones['palm.02.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'YXZ'
    pbone = obj.pose.bones[bones['palm.03.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'YXZ'
    pbone = obj.pose.bones[bones['palm.04.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'YXZ'
    pbone = obj.pose.bones[bones['f_index.01.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    try:
        pbone.rigify_parameters.separate_extra_layers = True
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.extra_layers = [False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.tweak_extra_layers = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['thumb.01.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    try:
        pbone.rigify_parameters.extra_layers = [False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.separate_extra_layers = True
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.tweak_extra_layers = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['f_middle.01.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    try:
        pbone.rigify_parameters.separate_extra_layers = True
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.extra_layers = [False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.tweak_extra_layers = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['f_ring.01.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    try:
        pbone.rigify_parameters.separate_extra_layers = True
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.extra_layers = [False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.tweak_extra_layers = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['f_pinky.01.L']]
    pbone.rigify_type = 'limbs.simple_tentacle'
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    try:
        pbone.rigify_parameters.separate_extra_layers = True
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.extra_layers = [False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False]
    except AttributeError:
        pass
    try:
        pbone.rigify_parameters.tweak_extra_layers = False
    except AttributeError:
        pass
    pbone = obj.pose.bones[bones['f_index.02.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['thumb.02.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['f_middle.02.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['f_ring.02.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['f_pinky.02.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['f_index.03.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['thumb.03.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['f_middle.03.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['f_ring.03.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'
    pbone = obj.pose.bones[bones['f_pinky.03.L']]
    pbone.rigify_type = ''
    pbone.lock_location = (False, False, False)
    pbone.lock_rotation = (False, False, False)
    pbone.lock_rotation_w = False
    pbone.lock_scale = (False, False, False)
    pbone.rotation_mode = 'QUATERNION'

    bpy.ops.object.mode_set(mode='EDIT')
    for bone in arm.edit_bones:
        bone.select = False
        bone.select_head = False
        bone.select_tail = False
    for b in bones:
        bone = arm.edit_bones[bones[b]]
        bone.select = True
        bone.select_head = True
        bone.select_tail = True
        arm.edit_bones.active = bone

    for eb in arm.edit_bones:
        if ('arm' in eb.name) or ('hand' in eb.name):
            eb.layers = (False, False, False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False)
        else:
            eb.layers = (False, False, False, False, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False)
    arm.layers = (False, False, False, False, False, True, False, True, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False, False)


if __name__ == "__main__":
    create_sample(bpy.context.active_object)